from typing import List, Dict, Tuple

STOP_TOKEN_DEFAULT = "<|stop|>"  # placement for stop token


def _minimal_reversible_patch(
    tokens: List[int],
    center_idx: int,
    tokenizer,
) -> Tuple[int, int]:
    """
    Return the *smallest* contiguous slice [start, end) that

    • contains ``center_idx`` and
    • round‑trips through decode‑then‑encode unchanged, i.e.
      ``tokens[start:end] == tokenizer.encode(
           tokenizer.decode(tokens[start:end], skip_special_tokens=True),
           add_special_tokens=False
      )``

    This is the formal definition of a *patch* in the docstring.
    """
    if isinstance(tokens, str):
        tokens = tokenizer.encode(tokens, add_special_tokens=False)
    n = len(tokens)
    # First try the trivial 1‑token span; for most Latin text it succeeds.
    start, end = center_idx, center_idx + 1

    def reversible(s: int, e: int) -> bool:
        sub = tokens[s:e]
        text = tokenizer.decode(sub, skip_special_tokens=True)
        return tokenizer.encode(text, add_special_tokens=False) == sub

    if reversible(start, end):
        return start, end
    for token_num in range(1, n + 1):
        for bias in range(token_num):
            start = center_idx - bias
            end = start + token_num
            if start >= 0 and end <= n:
                if reversible(start, end):
                    return start, end

    # Fallback: the whole sequence (should never happen in normal text).
    return 0, n


def compute_token_level_supervision(
    *,
    rejected_content: str,
    chosen_content: str,
    tokenizer=None,
    STOP_TOKEN=STOP_TOKEN_DEFAULT,
) -> Dict:
    """
    Compute token‑level supervision signals.

    Parameters
    ----------
    chosen_content / rejected_content : str
        The two candidate strings – *chosen* is forked from *rejected*.
    tokenizer : PreTrainedTokenizerBase, optional

    Returns
    -------
    dict  with keys

        fork_token_idx : int
            Index of the first diverging token position.
        chosen_token_id / rejected_token_id : int
            Token‑IDs at that fork position.
        chosen_content / rejected_content : list[dict]
            Each list is ``[pre_patch, patch, post_patch]`` where every
            element is a dict ``{"tokens": List[int], "ignore_loss": bool}``.
            • For *chosen*, the patch always receives loss (``ignore_loss=False``).
            • For *rejected*, the patch receives rejected loss **only if**
              it is a *single* token.  When the patch spans > 1 token we set ``ignore_loss=True`` for
              safety, because a many‑token rejected is harmful.
    """
    from transformers import AutoTokenizer

    if tokenizer is None:
        tokenizer = UnicodeTokenizer()

    tok_ch = tokenizer.encode(chosen_content, add_special_tokens=False)
    tok_rj = tokenizer.encode(rejected_content, add_special_tokens=False)

    # 1️⃣ Find the fork position.
    max_common = min(len(tok_ch), len(tok_rj))
    fork_idx = next(
        (i for i in range(max_common) if tok_ch[i] != tok_rj[i]), max_common
    )
    # assert not (
    #     max_common == fork_idx and len(tok_ch) == len(tok_rj)
    # ), f"Chosen and rejected is same! {chosen_content} == {rejected_content}"
    # warning
    if max_common == fork_idx and len(tok_ch) == len(tok_rj):
        print(
            f"Warning: Chosen and rejected are the same! \n  chosen:...{chosen_content[-35:]}\nrejected:...{rejected_content[-35:]}"
        )

    chosen_token_id = tok_ch[fork_idx] if fork_idx < len(tok_ch) else STOP_TOKEN
    rejected_token_id = tok_rj[fork_idx] if fork_idx < len(tok_rj) else STOP_TOKEN

    # 2️⃣ Locate minimal reversible patches around the diverging tokens.
    ch_s, ch_e = _minimal_reversible_patch(tok_ch, fork_idx, tokenizer)
    rj_s, rj_e = _minimal_reversible_patch(tok_rj, fork_idx, tokenizer)

    def build_chunks(
        tokens: List[int],
        s: int,
        e: int,
        ignore_patch_loss: bool = False,
        rejected_loss=None,
        is_fork_on_stop=False,
    ):
        """
        Helper – split *tokens* into three segments and label whether
        cross‑entropy loss should be computed (``ignore_loss=False``) or
        masked away (``ignore_loss=True``).
        """
        pre = {
            "type": "text",
            "text": tokenizer.decode(tokens[:s], skip_special_tokens=True),
            "ignore_loss": True,
        }
        patch = {
            "type": "text",
            "text": tokenizer.decode(tokens[s:e], skip_special_tokens=True),
            "ignore_loss": ignore_patch_loss,
            "tokens": tokens[s:e],
        }
        if rejected_loss:
            patch["rejected_loss"] = True
        post = {
            "type": "text",
            "text": tokenizer.decode(tokens[e:], skip_special_tokens=True),
            "ignore_loss": True,
        }
        if is_fork_on_stop:
            return [pre, patch]
        return [pre, patch, post]

    chosen_chunks = build_chunks(
        tok_ch, ch_s, ch_e, is_fork_on_stop=chosen_token_id == STOP_TOKEN
    )
    # Only apply negative loss when the rejected patch is exactly one token.
    rejected_chunks = build_chunks(
        tok_rj,
        rj_s,
        rj_e,
        ignore_patch_loss=(rj_e - rj_s != 1),
        rejected_loss=True,
        is_fork_on_stop=rejected_token_id == STOP_TOKEN,
    )
    # set chosen_text and rejected_text
    chosen_text = next(
        chunk for chunk in chosen_chunks if not chunk.get("ignore_loss")
    )["text"]
    rejected_text = next(
        chunk for chunk in rejected_chunks if chunk.get("rejected_loss")
    )["text"]
    return {
        "fork_token_idx": fork_idx,
        "chosen_token_id": chosen_token_id,
        "rejected_token_id": rejected_token_id,
        "chosen_text": chosen_text,
        "rejected_text": rejected_text,
        "chosen_text_unicode_range": [
            len(chosen_chunks[0]["text"]),
            len(chosen_chunks[1]["text"]),
        ],
        "rejected_text_unicode_range": [
            len(rejected_chunks[0]["text"]),
            len(rejected_chunks[1]["text"]),
        ],
        "chosen_content": chosen_chunks,
        "rejected_content": rejected_chunks,
    }


def apply_ignore_unicode_loss_mask_to_content(mask, content_str):
    previous_ignore_state = mask[0]
    previous_end_idx = 0
    content_patchs = []
    for idx, state in enumerate(list(mask) + ["add last patch finally"]):
        if state != previous_ignore_state:
            patch = dict(
                text=content_str[previous_end_idx:idx],
                ignore_loss=previous_ignore_state,
                type="text",
            )
            content_patchs.append(patch)
            previous_end_idx = idx
            previous_ignore_state = state
    return content_patchs


class UnicodeTokenizer:
    def __init__(self):
        self.name_or_path = "UnicodeTokenizer"

    def encode(self, string, **kwargs):
        return [ord(c) for c in list(string)]

    def decode(self, tokens, **kwargs):
        return "".join([chr(i) for i in tokens])

    def apply_chat_template(self, messages, tokenize=True, **kwargs):
        import json

        chatml = json.dumps(messages, indent=2, ensure_ascii=False)
        if tokenize:
            return self.encode(chatml)
        return chatml


unicode_tokenizer = UnicodeTokenizer()

# ----------------------------------------------------------------------
# ------------------------------ TESTS ---------------------------------
# ----------------------------------------------------------------------


def _patch_len(result: Dict, which: str, tokenizer) -> int:
    """Utility to grab the token length of the second chunk (the patch)."""
    return len(
        tokenizer.encode(
            result[f"{which}_content"][1]["text"], add_special_tokens=False
        )
    )


def test_one_token_align_one_patch(tok):
    """Patches are single tokens."""
    res = compute_token_level_supervision(
        chosen_content="I love cats.", rejected_content="I love dogs.", tokenizer=tok
    )
    assert _patch_len(res, "chosen", tok) == 1
    assert _patch_len(res, "rejected", tok) == 1
    # Negative loss **is** applied to the rejected patch.

    assert res["rejected_content"][1]["ignore_loss"] is False
    g()

    res = compute_token_level_supervision(
        chosen_content="Answer is中国",
        rejected_content="Answer is Chinese",
        tokenizer=tok,
    )
    g()
    # assert len(res["chosen_content"][1]["text"]) > 1


def test_many_token_align_one_patch(tok):
    """Chosen patch >1 token, rejected patch 1 token."""
    res = compute_token_level_supervision(
        chosen_content="I like '🥢'", rejected_content="I like '🥄'", tokenizer=tok
    )
    g()
    assert _patch_len(res, "chosen", tok) > 1
    assert _patch_len(res, "rejected", tok) > 1
    # Negative loss still applies (rejected patch single token).
    assert res["rejected_content"][1]["ignore_loss"] is True


def test_many_token_align_many_patch(tok):
    """' 🥢' == [11162, 98, 95]"""
    res = compute_token_level_supervision(
        chosen_content="prefix 🥢subfix",  # with space
        rejected_content="prefix🥢subfix",  # without space
        tokenizer=tok,
    )
    rejected_patch = res["rejected_content"][1]
    g()
    if len(rejected_patch["tokens"]) > 1:
        assert rejected_patch[
            "ignore_loss"
        ], "if rejected_patch include many tokens, ignore loss. Because a many‑token rejected loss is harmful."


def test_fork_token_is_stop_token(tok):
    chosen_stop_res = compute_token_level_supervision(
        chosen_content="prefix",
        rejected_content="prefix subfix",
        tokenizer=tok,
    )
    rejected_stop_res = compute_token_level_supervision(
        chosen_content="prefix continue",
        rejected_content="prefix",
        tokenizer=tok,
    )
    g()


def test_fork_token_is_last_token(tok):
    last_token_res = compute_token_level_supervision(
        chosen_content="1 2 3",
        rejected_content="1 2 4",
        tokenizer=tok,
    )
    chosen_content = last_token_res["chosen_content"]
    g()
    assert (
        len(chosen_content) == 3
        and chosen_content[-1]["text"] == ""
        and chosen_content[-1]["ignore_loss"]
    ), "The stop token is last patch with empty string. Should ignore it's loss"


if __name__ == "__main__":
    from boxx import *
    from test_utils import build_test_tokenizer

    tokenizer = tok = build_test_tokenizer()
    # tokenizer = UnicodeTokenizer()
    test_one_token_align_one_patch(tokenizer)
    # test_many_token_align_one_patch(tokenizer)
    # test_one_to_many(tokenizer)
    test_many_token_align_many_patch(tokenizer)
    test_fork_token_is_stop_token(tokenizer)
    test_fork_token_is_last_token(tokenizer)



import random

judge_message_default_cn = dict(
    role="system",
    content="现在你的身份变成了一个评判者，要对答复进行对比和评价。上面的最后一轮有上下两条答复，请分析和对比两条答复在对应语境和设定下的质量，最后选出一个最佳答复，允许平手。分析和评价完成后，需要把最优的答复以 `answer{}` 格式包裹，只允许填写 上、平、下， 比如 `answer{平}`",
)
custom_single_choice_default_cn = {
    "name": "谁更好",
    "type": "single_choice",
    "single_choice": [
        {"k": "上", "v": None},
        {"k": "平", "v": None},
        {"k": "下", "v": None},
    ],
    "tips": "最后一轮有上下两条答复, 谁更好? 允许平手",
    "required": True,
}


def build_panda_battle(
    arena_json1, arena_json2, judge_message=None, custom_single_choice=None
):
    if judge_message is None:
        judge_message = judge_message_default_cn
    if custom_single_choice is None:
        custom_single_choice = custom_single_choice_default_cn
    if "data" in arena_json1 and isinstance(arena_json1["data"], dict):
        arena_json1 = arena_json1["data"]
    if "data" in arena_json2 and isinstance(arena_json2["data"], dict):
        arena_json2 = arena_json2["data"]
    panda_list = []
    for prompt_key in arena_json1:
        if prompt_key not in arena_json2:
            continue
        msgs1 = arena_json1[prompt_key]
        msgs2 = arena_json2[prompt_key]
        is_switched = random.random() < 0.5
        if is_switched:
            up, down = msgs2[-1], msgs1[-1]
            exp_up = up["arena"]["exp_name"]
            exp_down = down["arena"]["exp_name"]
        else:
            up, down = msgs1[-1], msgs2[-1]
            exp_up = up["arena"]["exp_name"]
            exp_down = down["arena"]["exp_name"]

        hidden_info = dict(
            order=[exp_up, exp_down], is_switched=is_switched, prompt_key=prompt_key
        )
        judge_messages = msgs1[:-1] + [up, down, judge_message]
        panda = {
            "dialogs": {
                "1": {
                    "hidden_info": hidden_info,
                    "messages": judge_messages,
                    "annotate": {
                        "is_good": None,
                        "customs": [custom_single_choice],
                    },
                }
            },
            "description": f"prompt_name:{msgs1[-1]['arena'].get('prompt_name', prompt_key)}",
        }
        panda_list.append(panda)
    return panda_list


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  5 21:19:59 2025

@author: DIYer22
"""
import mxlm
import mximport
from copy import deepcopy

with mximport.inpkg():
    from ..token_level_supervision_utils import unicode_tokenizer

correcting_sft_system_prompt_cn = """- 先前的 system prompt 只做评估用，不必再遵守
- 你本体是一个 GPT 架构的 LLM, 你现在的角色切换为了 token-level correcting model
- 目标是通过修改不恰当的 token 来优化已有的回答
- 你的任务是：
    1. 定位上述回答中，第一个不恰当的 token，即指出 “修改位置”
    2. 将“不恰当 token”修改为更加恰当的 token，使得基于 “恰当 token” 继续做补全能获得最好、最准确的答复
- Correcting 范围：多轮的情况下，只定位和修改上一轮（即最新轮）的答复中首个“不恰当 token”
- 由于你作为 LLM 只会输出文本，我们按照这个文本格式来输出你的 correcting 答复:
    - `<|split|>{location_tokens}<|split|>{location_index}<|split|>{replacement_token}<|split|>`
    - `<|split|>` 是分隔内容的 special token，且回答必须以 `<|split|>` 作为开头和结尾
    - `{location_tokens}`: 用来定位 “修改位置” 的一串 tokens
        - 其内容为从不恰当的 token 开始，持续生成，直到触发以下任意情况：
            1. 在所有模型输出的 tokens 中 (包括模型的历史输出) 被 `{location_tokens}` 匹配上的第一处位置正好就是 “修改位置” 
                - 此时的 `{location_index}` 应该为 0，并停止生成
                - 若第一匹配处不是 “修改位置”，则继续生成下一个 token 来做更加精准的定位
            2. `{location_tokens}` 长度达到 20 个 token，就该停止生成了
                - 但是，若最后的几个 token 不能被你自己 (correcting model) 的 tokenizer decode 为完整字符，需要突破 20 tokens 限制生成到能 decode 出完整字符为止
                - 若 20 个 token 都没法把 “修改位置” 准确定位，那就需要配合 `{location_index}` 来一起定位了
            3. 一轮结束了，即已经生成了 stop token: `<|stop|>`，也应该停止生成
    - `{location_index}` 表示在所有模型输出的 tokens 中, 能被 `{location_tokens}` 匹配上的所有位置中的第几个位置
        - 是一个 int 数值，从 0 开始计数，支持负数，和 Python list 的 index 一致
        - 当用负数表示 index 时的绝对值比正数 index 更加小的时候，`{location_index}` 就用负数表示
        - `{location_tokens}` 和 `{location_index}` 配合后，能在所有答复中共同定位一个唯一的位置，即 “第一个不恰当 token” 的位置
    - `{replacement_token}`: 更加恰当的 token，期望改为恰当 token 后，继续做补全能获得最好、最准确的答复。只需要一个 token 即可
    -  stop token: 上面的每一轮答复最后都有 stop token，需要的话，在 `{location_tokens}`,`{replacement_token}` 中使用 special token `<|stop|>` 来表示 stop token
        - 比如, 要续写最后一轮的答复 `<|split|><|stop|><|split|>-1<|split|>{continue token}<|split|>`
    - tokenizer 问题：
        - 你需要通过多输出 token 或提前输出 token 来避免潜在的 tokenizer decode 出不合规文本的问题。
        - 即多个 tokens 对应一个文本字符的情况下，要把多个 token 视为一个整体，使所有输出的 tokens 能和文本互相转换，而不要截断中间 token
    - 如果 Correcting 范围内的回答都没有问题，输出 `<|split|><|split|>`

## example 1:
USER:
列举 3 种水果：
ASSISTANT:
苹果、土豆、香蕉
期望的输出: “<|split|>土豆<|split|>0<|split|>西瓜<|split|>”

## example 2:
USER:
Just reply 2 times, Using "|" as a separator:
1;2;3;4;5;6;7;8;9;8;
ASSISTANT:
1;2;3;4;5;6;7;8;9;8;|1;2;3;4;5;6;7;8;9;8;
USER:
Reply again
ASSISTANT:
1;2;3;4;5;6;7;8;9;8;|1;2;3;4;5;6;7;8;9;8;|1;2;3;4;5;6;7;8;9;8;

期望的输出: “<|split|>|1;2;3;4;5;6;7;8;9;8<|split|>-1<|split|><|stop|><|split|>”
- “第一个不恰当 token”处和其他 ASSISTANT 的回答有重复，所以会生成完整个 20 个 `{location_tokens}`
- `{location_index}` 用正数表示时为 2， 用负数为 -1，其中， -1 绝对值更加小，所以应该用 -1
- 此处 `{replacement_token}` 为 stop token"""

correcting_sft_system_prompt_default = correcting_sft_system_prompt_cn


def next_decodable_num(tokens, current_num, tokenizer):
    """
    从 tokens 的 current_num 位置开始，找到下一个能被 tokenizer decode 出完整字符的 idx
    """
    for num in range(current_num + 1, len(tokens) + 1):  # number of tokens
        try:
            decoded_text = tokenizer.decode(tokens[0:num])
            if (
                tokenizer.encode(decoded_text, add_special_tokens=False)
                == tokens[0:num]
            ):
                return dict(next_num=num, decoded_text=decoded_text)
        except Exception:
            continue
    raise ValueError(
        "无法找到下一个可解码的位置",
        getattr(tokenizer, "name_or_path", "unknow_tokenizer"),
        tokens,
    )


class NextTokenPredictionAsCorrectingBuilder:
    def __init__(
        self,
        tokenizer=None,
        SPLIT_TOKEN="<|split|>",  # for qwen 2.5
        STOP_TOKEN="<|stop|>",
        max_location_tokens=20,
        scope_slice=(-1, None),  # TODO: slice of which messages can be correcting
    ):
        self.tokenizer = tokenizer or unicode_tokenizer
        self.SPLIT_TOKEN = SPLIT_TOKEN
        self.STOP_TOKEN = STOP_TOKEN
        self.max_location_tokens = max_location_tokens
        self.scope_slice = scope_slice

    def get_correcting_sft_system_prompt(self, language="cn"):
        if language == "cn":
            prompt = correcting_sft_system_prompt_cn
        else:
            prompt = correcting_sft_system_prompt_default
        return (
            prompt.replace("<|split|>", self.SPLIT_TOKEN)
            .replace("<|stop|>", self.STOP_TOKEN)
            .replace(" 20 ", f" {self.max_location_tokens} ")
        )

    def convert_token_level_to_unicode_location(self, rejected_msgs):
        """
        根据 rejected_msgs 中的 token_level 信息返回 unicode_location

        Args:
            rejected_msgs: 消息列表

        Returns:
            dict: {"message_index": int, "unicode_index": int}
        """
        # 查找首个有 token_level 的 assistant 消息
        for i, msg in enumerate(rejected_msgs):
            if msg["role"] == "assistant" and "token_level" in msg:
                token_level = msg["token_level"]
                unicode_location = token_level["rejected_text_unicode_range"][0]
                return {"message_index": i, "unicode_index": unicode_location}
        return {"not_found": True}

    def parser_ntp_as_correcting_text(self, ntp_as_correcting_text):
        mid_text = ntp_as_correcting_text.removeprefix(self.SPLIT_TOKEN).removesuffix(
            self.SPLIT_TOKEN
        )
        if mid_text:  # correcting
            splits = mid_text.split(self.SPLIT_TOKEN)
            # TODO: How to handle exception?
            assert len(splits) == 3, splits
            ntp_as_correcting = dict(
                zip(["location_text", "location_index", "replacement_token"], splits)
            )
            ntp_as_correcting["location_index"] = int(
                ntp_as_correcting["location_index"]
            )
        else:  # is_good
            ntp_as_correcting = dict(is_good=True, location_text="")
        return ntp_as_correcting

    def get_unicode_location(self, msgs, ntp_as_location=None):
        """
        根据 ntp_as_location 定位 unicode_location
        如果 ntp_as_location is None, 则从 msgs 必须是 correcting_sft, 会从最后一条消息中解析出 ntp_as_location
        """
        if ntp_as_location is None:
            sys_msg, correcting_msg = msgs[-2:]
            msgs = msgs[:-2]
            # ntp_as_correcting_gt = correcting_msg.get('correcting')
            ntp_as_correcting_text = mxlm.get_text_content(correcting_msg)
            ntp_as_location = self.parser_ntp_as_correcting_text(ntp_as_correcting_text)
            if ntp_as_location.get("is_good"):
                return dict(not_found=True, is_good=True)
        unicode_location = self._get_unicode_location(msgs, ntp_as_location)
        return unicode_location

    def _get_unicode_location(self, msgs, ntp_as_location):
        """
        Compute unicode_location by ntp_as_location in messages without token_level_info
        if Not found:
            return dict(not_found=True)

        用 for loop 遍历所有 assistant 消息，找到所有能匹配上 location_text 的位置
        如果能找到， 返回 location_index 对应的位置的 unicode_location
        否则返回 not_found=True
        """
        unicode_sequence_dic = self.messages_to_assistant_unicode_sequence(msgs)
        assistant_sequence = unicode_sequence_dic["assistant_sequence"]
        location_index = ntp_as_location["location_index"]
        location_text = ntp_as_location.get("location_text", "")
        assert location_text, ntp_as_location
        unicode_locations = []
        for message_index, assistant_content in zip(
            unicode_sequence_dic["assistant_indices"],
            assistant_sequence.split(self.STOP_TOKEN),
        ):

            assistant_content += self.STOP_TOKEN
            start = 0
            while True:
                index = assistant_content.find(location_text, start)
                if index == -1:
                    break
                unicode_location = dict(
                    message_index=message_index, unicode_index=index
                )
                unicode_locations.append(unicode_location)
                start = index + 1
        match_num = len(unicode_locations)

        if match_num and -match_num <= location_index and location_index < match_num:
            unicode_location = unicode_locations[location_index]
            unicode_location["match_num"] = match_num
            return unicode_location
        else:
            return dict(not_found=True, match_num=match_num)

    def messages_to_assistant_unicode_sequence(self, msgs, unicode_location=None):
        """
        Convert messages to a single text sequence, if unicode_location is given,
        also compute the sequence_index in the combined text sequence.

        Returns:
            update to unicode_location dict: {"assistant_sequence": str, "sequence_index": int (if unicode_location is given)}
        """

        # 收集所有assistant消息的内容，并记录其在原始消息中的索引
        assistant_contents = []
        assistant_indices = []
        for i, msg in enumerate(msgs):
            if msg["role"] == "assistant":
                content = mxlm.get_text_content(msg["content"])
                # 添加隐藏的 STOP_TOKEN
                content += self.STOP_TOKEN
                # content += "\n\n-----\n\n" 会导致潜在的 tokenizer 粘连问题
                assistant_contents.append(content)
                assistant_indices.append(i)

        assistant_sequence = "".join(assistant_contents)
        if unicode_location is None:
            unicode_location = {}
        else:
            message_index = unicode_location["message_index"]
            target_unicode_index = unicode_location["unicode_index"]
            # 计算目标位置的unicode位置
            # 找到目标消息在assistant消息列表中的索引
            try:
                assistant_msg_idx = assistant_indices.index(message_index)
            except ValueError:
                raise ValueError(f"消息索引 {message_index} 不是 assistant 消息")

            current_index = 0
            for i in range(assistant_msg_idx):
                current_index += len(assistant_contents[i])
            sequence_index = current_index + target_unicode_index
            # unicode_location = deepcopy(unicode_location)
            unicode_location["sequence_index"] = sequence_index
        unicode_location["assistant_sequence"] = assistant_sequence
        unicode_location["assistant_indices"] = assistant_indices
        # print(unicode_location)
        return unicode_location

    def set_location_index(self, rejected_msgs, ntp_as_location, unicode_location):
        """
        在所有模型输出的 tokens 中查找 ntp_as_location.location_text 的所有匹配位置，
        返回对应的索引位置 ntp_as_location.location_index

        Args:
            rejected_msgs: 消息列表
            ntp_as_location: dict(location_text=...) or 要查找的字符串
            unicode_location: dict, 包含 message_index 和 unicode_index, 也可以包含 assistant_sequence 和 sequence_index

        Returns ntp_as_location:
            int: location_index，从0开始计数，负数表示从末尾倒数
        """
        if isinstance(ntp_as_location, str):
            ntp_as_location = dict(location_text=ntp_as_location)
        ntp_as_location = deepcopy(ntp_as_location)
        location_text = ntp_as_location["location_text"]
        if "assistant_sequence" not in unicode_location:
            unicode_location = self.messages_to_assistant_unicode_sequence(
                rejected_msgs, unicode_location
            )
        assistant_sequence = unicode_location["assistant_sequence"]
        sequence_index = unicode_location["sequence_index"]

        # 在所有assistant内容中查找location_text的所有匹配位置
        matches = []
        start = 0
        while True:
            index = assistant_sequence.find(location_text, start)
            if index == -1:
                break
            matches.append(index)
            start = index + 1

        location_index = None
        # 找到目标位置对应的匹配索引
        for idx, match_index in enumerate(matches):
            if match_index == sequence_index:
                # 如果负数的绝对值更小，使用负数表示
                negative_idx = idx - len(matches)
                if abs(negative_idx) < idx:
                    location_index = negative_idx
                else:
                    location_index = idx

        ntp_as_location.update(
            unicode_location=unicode_location, match_num=len(matches)
        )
        ntp_as_location["location_index"] = location_index
        if not len(matches):
            ntp_as_location["not_found"] = True
        return ntp_as_location

    def convert_rejected_content_to_ntp_as_location(self, rejected_msgs):
        """
        将 rejected_msgs 和 token_level_info 转换为 Next Token Prediction as location 格式

        - 获得 correcting 位置的 unicode_location
        - 从 unicode_location 处取 suffix 再 decode
        - 循环 next valid decodable 直到 location_index 为 0，或者 token 超长
        - 生成并返回 location_text 和 location_index

        Args:
            rejected_msgs: 消息列表

        Returns:
            dict: {"location_text": str, "location_index": int}
        """
        # 获取 unicode_location
        unicode_location = self.convert_token_level_to_unicode_location(rejected_msgs)
        message_index = unicode_location["message_index"]
        unicode_index = unicode_location["unicode_index"]

        content = mxlm.get_text_content(rejected_msgs[message_index]["content"])
        content_suffix = content[unicode_index:] + self.STOP_TOKEN
        suffix_tokens = self.tokenizer.encode(content_suffix, add_special_tokens=False)
        decodable_num = 0

        while True:
            decodable_res = next_decodable_num(
                suffix_tokens, decodable_num, self.tokenizer
            )
            decodable_num = decodable_res["next_num"]
            location_text = decodable_res["decoded_text"]
            ntp_as_location = self.set_location_index(
                rejected_msgs,
                location_text,
                unicode_location,
            )
            if ntp_as_location.get("not_found"):
                raise ValueError("无法定位到 location_text", ntp_as_location)
            location_index = ntp_as_location.get("location_index", None)
            if location_index == 0:
                break
            if decodable_num >= len(suffix_tokens):
                break
            if decodable_num >= self.max_location_tokens:
                break

        ntp_as_location["location_tokens"] = suffix_tokens[:decodable_num]

        if "asset_location_consistency":
            unicode_location2 = self.get_unicode_location(
                rejected_msgs, ntp_as_location
            )
            assert (
                unicode_location["message_index"] == unicode_location2["message_index"]
                and unicode_location["unicode_index"]
                == unicode_location2["unicode_index"]
            ), (
                "asset_location_consistency: "
                + str(unicode_location)
                + str(unicode_location2)
                + str(ntp_as_location)
            )
        return ntp_as_location

    def build_correcting_sft_by_token_level_SFT(
        self, msgs, is_good=None
    ):  # must be is_good SFT msgs or token_level_SFT msgs
        unicode_location = self.convert_token_level_to_unicode_location(msgs)

        sys_prompt_message = dict(
            role="system",
            content=self.get_correcting_sft_system_prompt(),
        )
        # double check
        if is_good is not None:
            assert bool(is_good) == bool(
                unicode_location.get("not_found")
            ), f"is_good must consistent with token_level_info, is_good: {is_good} != unicode_location: {unicode_location}"

        [msg.update(ignore_loss=True) for msg in msgs if msg["role"] == "assistant"]
        if unicode_location.get(
            "not_found"
        ):  # 没有 token_level 信息, 属于 is_good 的 SFT
            is_good_correcting_msg = dict(
                role="assistant",
                content=self.SPLIT_TOKEN * 2,
                correcting=dict(is_good=True, scope_slice=self.scope_slice),
            )
            correcting_sft = msgs + [sys_prompt_message, is_good_correcting_msg]

            return correcting_sft
        else:  # 有 token_level 信息, 属于 not is_good 的 token-level SFT
            token_level_msg = msgs[-1]
            token_level_info = token_level_msg["token_level"]
            rejected_content_chunks = token_level_info.pop("rejected_content")
            token_level_info["chosen_content"] = token_level_msg["content"]

            rejected_content_str = mxlm.get_text_content(rejected_content_chunks)
            rejected_msg = dict(
                role="assistant",
                ignore_loss=True,
                content=rejected_content_str,
                finish_reason=token_level_info.get("rejected_finish_reason", ""),
                token_level=token_level_info,
            )
            rejected_msgs = msgs[:-1] + [rejected_msg]

            ntp_as_location = self.convert_rejected_content_to_ntp_as_location(
                rejected_msgs,
            )
            ntp_as_correcting = deepcopy(ntp_as_location)
            ntp_as_correcting.pop("unicode_location", None)
            replacement_text = (
                token_level_info["chosen_text"] or self.STOP_TOKEN
            )  # if chosen_text is empty mean chosen stop token
            ntp_as_correcting.update(
                replacement_text=replacement_text,
                is_good=False,
                scope_slice=self.scope_slice,
            )

            correcting_content = f"{self.SPLIT_TOKEN}{ntp_as_correcting['location_text']}{self.SPLIT_TOKEN}{ntp_as_correcting['location_index']}{self.SPLIT_TOKEN}{ntp_as_correcting['replacement_text']}{self.SPLIT_TOKEN}"
            correcting_msg = dict(
                role="assistant",
                content=correcting_content,
                correcting=ntp_as_correcting,
            )
            correcting_sft = rejected_msgs + [
                sys_prompt_message,
                correcting_msg,
            ]
        # import boxx.g
        return correcting_sft

    def apply_ntp_as_correcting(self, msgs, ntp_as_correcting: str | dict):
        if isinstance(ntp_as_correcting, str):
            ntp_as_correcting = self.parser_ntp_as_correcting_text(ntp_as_correcting)
        if ntp_as_correcting.get("is_good"):
            return dict(
                ntp_as_correcting=ntp_as_correcting,
            )
        unicode_location = self.get_unicode_location(msgs, ntp_as_correcting)
        if unicode_location.get("not_found"):
            return dict(
                ntp_as_correcting=ntp_as_correcting, unicode_location=unicode_location
            )
        else:
            msg_idx = unicode_location["message_index"]
            partial_msg = deepcopy(msgs[msg_idx])
            if isinstance(partial_msg["content"], list):
                assert all(
                    [d["type"] == "text" for d in partial_msg["content"]]
                ), partial_msg
                partial_msg["content"] = mxlm.get_text_content(partial_msg["content"])
            good_prefix = partial_msg["content"][: unicode_location["unicode_index"]]
            if self.STOP_TOKEN == ntp_as_correcting["replacement_token"]:
                # no need continue final message
                partial_msg["content"] = good_prefix
                partial_msg["finish_reason"] = "stop"
            else:
                partial_msg["content"] = (
                    good_prefix + ntp_as_correcting["replacement_token"]
                )
                if "finish_reason" in partial_msg:
                    del partial_msg["finish_reason"]
            partial_messages = msgs[:msg_idx] + [partial_msg]
            correction = dict(
                ntp_as_correcting=ntp_as_correcting,
                unicode_location=unicode_location,
                partial_messages=partial_messages,
            )
            return correction


if __name__ == "__main__":
    from boxx import *

    with mximport.inpkg():
        from ..test_utils import build_test_tokenizer, get_test_rejected_msgs1
        from ..parser import build_test_panda_tree

    panda_json_dir = "../../../on-panda-example-data/panda_json"
    tokenizer = build_test_tokenizer()
    # build_argkws = dict(tokenizer=unicode_tokenizer)
    build_argkws = dict(
        tokenizer=tokenizer,
        SPLIT_TOKEN="<|fim_pad|>",  # for qwen 2.5
        STOP_TOKEN="<|fim_suffix|>",
    )
    builder = NextTokenPredictionAsCorrectingBuilder(**build_argkws)

    # test next_decodable_num
    complex_emoji_text = "🧎🏿‍♂️‍➡️"
    decodable = next_decodable_num(tokenizer.encode(complex_emoji_text), 0, tokenizer)
    assert decodable["next_num"] != 1, decodable

    # test sample case
    rejected_msgs1, ntp_as_correcting_text_gt1 = get_test_rejected_msgs1()[:2]

    result1 = builder.convert_rejected_content_to_ntp_as_location(rejected_msgs1)
    assert result1["location_text"] == " potato", result1
    assert result1["location_index"] == 0, result1

    correction1 = builder.apply_ntp_as_correcting(
        rejected_msgs1, ntp_as_correcting_text_gt1
    )
    assert correction1["partial_messages"][-1]["content"] == "Apple, orange"
    assert (
        "finish_reason" not in correction1["partial_messages"][-1]
    ), "Should continue_final_message (no finish_reason)"

    # test correcting_sft extreme cases: chosen stop
    test_json = (
        f"{panda_json_dir}/2025-09-10_correcting_sft_tokenizer-Qwen2.5.panda.json"
    )
    panda_tree = build_test_panda_tree(test_json)
    correcting_sft2 = panda_tree.build_correcting_sft_data_v1(builder)[-1]
    correcting_content2 = correcting_sft2[-1]["content"]
    ntp_as_correcting_text_gt2 = "<|fim_pad|>|1;2;3;4;5;6;7;8;9;8<|fim_pad|>-1<|fim_pad|><|fim_suffix|><|fim_pad|>"
    assert correcting_content2 == ntp_as_correcting_text_gt2, correcting_content2
    correction2 = builder.apply_ntp_as_correcting(
        correcting_sft2[:-2], ntp_as_correcting_text_gt2
    )
    assert correction2["partial_messages"][-1]["finish_reason"] == "stop"

    # test correcting_sft extreme cases: chosen continue
    test_json3 = f"{panda_json_dir}/2025-09-11_correcting_sft_continue_tokenizer-Qwen2.5.panda.json"
    panda_tree3 = build_test_panda_tree(test_json3)
    correcting_sft3 = panda_tree3.build_correcting_sft_data_v1(builder)[-1]
    correcting_content3 = correcting_sft3[-1]["content"]
    assert (
        correcting_content3
        == "<|fim_pad|><|fim_suffix|><|fim_pad|>1<|fim_pad|>|<|fim_pad|>"
    ), correcting_content3

    # test single_char_repeat case: chosen stop
    test_json4 = (
        f"{panda_json_dir}/2025-09-12_single_char_repeat_tokenizer-Qwen2.5.panda.json"
    )
    panda_tree4 = build_test_panda_tree(test_json4)
    correcting_sft4 = panda_tree4.build_correcting_sft_data_v1(builder)[-1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 21 16:23:49 2024

@author: DIYer22
"""
true = True
false = False
null = None


def messages_to_panda_tree(msgs, uuid=None):
    panda_tree = {
        "dialogs": {
            "1": {
                "messages": msgs,
                "annotate": {
                    "is_good": None,  # Open Data Pannel to annotate
                },
            }
        }
    }
    if uuid is not None:
        panda_tree["uuid"] = uuid
    return panda_tree


if __name__ == "__main__":
    pass
    image_url_msg_example = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "图中左侧的 “v” 是由什么形状构成？\n"},
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "https://docs.vllm.ai/en/latest/_static/vllm-logo-text-light.png"
                    },
                },
            ],
        }
    ]
    panda_tree_example = {
        "dialogs": {
            "1": {
                "messages": [
                    {"role": "user", "content": "🍓How many `R` in strawberry?"}
                ],
                "operations": [
                    {
                        "time": 1734883835911,
                        "parent": "1",
                        "chat_config": {
                            "model": "vllm-model",
                            "max_tokens": 1024,
                            "temperature": 0.5,
                        },
                        "common_prefix_length": -1,
                        "operator": "generate_new",
                        "is_new_generated": true,
                        "on_policy": true,
                    }
                ],
            }
        },
        "version": "1.0",
        "uuid": "2024-12-23_00-10-35.908",
        "hash_map": {},
        "deleted_dialogs": {},
        "update_time": 1734883863852,
    }
    panda_tree = messages_to_panda_tree(image_url_msg_example)
    from pprint import pp

    pp(panda_tree)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 23 16:56:53 2025

@author: DIYer22
"""
import mximport

with mximport.inpkg():
    from .is_good_score_mixin import IsGoodScoreMixin


class TokenLevelCorrectingModelMeta:
    """
    TODO: using this or delete
    Meta class of different token-level correcting models. e.g.:
    - CorrectingCopyResponseModel: Using Copy response to get whole context for each generated token, and better computation for each token
    - CorrectingSftModel: Next Token Prediction as correcting location
    - Reasoning for Correcting using JSON output with right prefix
    - Bidirectional correcting head for location with whole context
    """

    def __init__(self):
        pass

    def correct(self, text: str) -> str:
        # Implement token-level correction logic here
        return text


class CorrectingSftModel(TokenLevelCorrectingModelMeta, IsGoodScoreMixin):
    def __init__(
        self,
        chat_correcting,
        sft_correcting_builder,
    ):
        self.chat_correcting = chat_correcting
        self.builder = sft_correcting_builder

    def build_correcting_prompt(self, msgs):

        sys_prompt_message = dict(
            role="system",
            content=self.builder.get_correcting_sft_system_prompt(),
        )
        return msgs + [sys_prompt_message]

    def generate_correction(self, msgs):
        """
        Input QA msgs, return unicode_location
        """
        correcting_prompt = self.build_correcting_prompt(msgs)
        response_dic = self.chat_correcting(
            correcting_prompt,
            return_dict=True,
            max_tokens=self.builder.max_location_tokens + 20,
        )
        ntp_as_correcting_text = response_dic["choices"][0]["message"]["content"]
        correction = self.builder.apply_ntp_as_correcting(msgs, ntp_as_correcting_text)
        return correction

    def generate_and_apply_correction(self, msgs, chat_policy):
        corrected = dict(model_policy=chat_policy.model)
        if msgs[-1]["role"] in ["assistant"]:
            corrected["model_correcting"] = self.chat_correcting.model
            correction = self.generate_correction(msgs)
            corrected["correction"] = correction
            if correction["ntp_as_correcting"].get("is_good"):
                corrected_msgs = msgs
            else:
                partial_messages = correction["partial_messages"]
                if partial_messages[-1].get("finish_reason") == "stop":
                    corrected_msgs = partial_messages
                else:
                    corrected_content = chat_policy(
                        partial_messages,
                        continue_final_message=True,
                        add_generation_prompt=False,
                        echo=True,
                    )
                    corrected_msgs = msgs[:-1] + [
                        dict(role="assistant", content=corrected_content)
                    ]
        else:  # make new message
            # corrected without model_correcting key means new message
            corrected_msgs = msgs + [dict(role="assistant", content=chat_policy(msgs))]

        corrected["corrected_messages"] = corrected_msgs
        # g()
        return corrected

    def correcting_sampling(self, msgs, chat_policy, n=5):
        correction_count = 0
        corrected_msgs = msgs
        correcteds = []
        for correction_idx in range(n):
            corrected = self.generate_and_apply_correction(corrected_msgs, chat_policy)
            correcteds.append(corrected.copy())
            corrected_msgs = corrected["corrected_messages"]
            if "correction" in corrected and corrected["correction"][
                "ntp_as_correcting"
            ].get("is_good"):
                break
            correction_count += 1
        corrected["correction_count"] = correction_count
        corrected["correction_count_max"] = n
        corrected["correcteds"] = correcteds
        return corrected


def build_test_correcting_sft_model(chat_correcting=None, builder=None):
    import mxlm
    import onpanda
    import transformers

    if chat_correcting is None:
        chat_correcting = mxlm.ChatAPI(
            model="step1f-correct-sft-it1200",
            temperature=0,
            top_p=1.0,
            max_tokens=40,
            logprobs=True,
            return_dict=True,
        )
    if builder is None:
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            "Qwen/Qwen2.5-7B-Instruct-GPTQ-Int4",
            use_fast=True,
            # local_files_only=True,
        )
        builder = onpanda.NextTokenPredictionAsCorrectingBuilder(
            tokenizer=tokenizer,
            SPLIT_TOKEN="<|fim_pad|>",  # for qwen 2.5
            STOP_TOKEN="<|fim_suffix|>",
        )
    return CorrectingSftModel(chat_correcting, builder)


if __name__ == "__main__":
    from boxx import *
    from onpanda.test_utils import get_test_rejected_msgs1
    from copy import deepcopy

    correct_model = build_test_correcting_sft_model()

    msgs = [
        {"role": "user", "content": "5+7="},
        {"role": "assistant", "content": "32"},
        # {"role": "assistant", "content": "12"},
    ]
    msgs = get_test_rejected_msgs1()[0]

    # msgs = [{"role": "user", "content": "How many `1` in result of 652*8596"},]

    chat_policy = deepcopy(correct_model.chat_correcting)
    chat_policy.default_kwargs["max_tokens"] = 1536
    corrected = correct_model.correcting_sampling(msgs, chat_policy, n=5)
    tree(corrected)



# -*- coding: utf-8 -*-

from .__info__ import __version__, __description__
from .utils import *
from .parser import *
from .arena.panda_battle import build_panda_battle
from .token_level_supervision_utils import (
    compute_token_level_supervision,
    unicode_tokenizer,
)
from .correcting_model.correcting_sft_utils import (
    NextTokenPredictionAsCorrectingBuilder,
    correcting_sft_system_prompt_cn,
    correcting_sft_system_prompt_default,
)
from .correcting_model.correcting_sft_model import CorrectingSftModel



import re
import mxlm
from copy import deepcopy
import mximport
import os
import json

with mximport.inpkg():
    from .token_level_supervision_utils import (
        compute_token_level_supervision,
        unicode_tokenizer,
        apply_ignore_unicode_loss_mask_to_content,
    )
    from .correcting_model.correcting_sft_utils import (
        NextTokenPredictionAsCorrectingBuilder,
    )

HASH_TEMPLATE_PREFIX = "<|hash|>"
HASH_TEMPLATE_REGEX = r"^<\|hash\|>([A-Za-z0-9+\/=]+)$"


def recover_hash_map(data):
    # work on any dialogs
    def recover(obj):
        if isinstance(obj, (dict, list)):
            items = obj.items() if isinstance(obj, dict) else enumerate(obj)
            for key, value in items:
                if isinstance(value, str) and re.match(HASH_TEMPLATE_REGEX, value):
                    hash_value = value.replace(HASH_TEMPLATE_PREFIX, "")
                    obj[key] = data["hash_map"][hash_value]
                elif isinstance(value, (list, dict)):
                    recover(value)

    recover(data["dialogs"])
    data["hash_map"] = {}
    return data


def sequence_prefix_length(seq1, seq2):
    for i in range(min(len(seq1), len(seq2))):
        if seq1[i] != seq2[i]:
            return i
    return i + 1


RESPONSE_ROLES = ["assistant"]


class PandaTreeParser:
    def __init__(self, tokenizer=None):
        self.tokenizer = tokenizer

    def parser(self, data):
        return PandaTree(data, tokenizer=self.tokenizer)


class PandaTree:
    SUPPORT_PANDA_TREE_VERSION = "2.0"

    def __init__(self, data, tokenizer=None):
        self.raw_data = data
        self.tokenizer = tokenizer
        self.data = data = self.pre_process(data)
        assert len(data["dialogs"]), "Empty dialogs: " + str(data)
        dialogs = data["dialogs"]
        dialog_valide_keys = sorted(dialogs)
        dense_keys = [
            k for k in dialog_valide_keys if dialogs[k]["annotate"].get("is_good")
        ]
        assert dense_keys, "No any is_good dialog in this data."

        trees = {}
        to_parent = {}

        def get_parents(dialog_key):  # include self
            parents = [dialog_key]
            while to_parent[parents[-1]]:
                parents.append(to_parent[parents[-1]])
            return parents[::-1]

        for dialog_key in dialog_valide_keys:
            dialog = data["dialogs"][dialog_key]
            operations = dialog.get("operations", [])
            is_tree_root = self.is_operation_tree_root(operations)
            if not is_tree_root:
                parent = int(operations[0]["parent"])
                if parent not in to_parent:
                    # belong to deleted
                    is_tree_root = True
            if is_tree_root:
                trees[dialog_key] = {}
                to_parent[dialog_key] = None
            else:
                to_parent[dialog_key] = parent
                parents = get_parents(parent)

                node = trees
                for _parent in parents:
                    node = node[_parent]
                node[dialog_key] = {}

        # best practice: only do negative supervision for outcome and fork pairs. because dense_keys will provide positive supervision. if do so, negative supervision should duplicate.
        # pair of (negative, positive)
        outcome_pairs = (
            []
        )  # pairs not include token level supervision. similar to DPO pair
        fork_pairs = []

        def flatten_tree(tre):
            if not tre:
                return []
            res = []
            for k in tre:
                res.append(k)
                res.extend(flatten_tree(tre[k]))
            return res

        for tree_key in trees:
            tre = trees[tree_key]  # avoid boxx.tree variable
            flattens = [tree_key] + flatten_tree(tre)

            tree_dense_keys = [
                dialog_key for dialog_key in flattens if dialog_key in dense_keys
            ]

            if tree_dense_keys:  # this tree has dense, only need to do fork pairs
                for dialog_key in flattens:
                    if dialog_key not in dense_keys:
                        # find in this tree's dense which has nearst sequence as pair dense
                        # may be multiple pair dense with the same prefix length
                        min_prefix_len = 9e10
                        pair_dense_keys = []
                        for dense_key in tree_dense_keys:
                            prefix_len = sequence_prefix_length(
                                dialogs[dialog_key]["sequence"],
                                dialogs[dense_key]["sequence"],
                            )
                            # for multiple pieces of data with the same branching point, token level negative supervision should duplicate
                            if prefix_len < min_prefix_len:
                                min_prefix_len = prefix_len
                                pair_dense_keys = [dense_key]
                            elif prefix_len == min_prefix_len:
                                pair_dense_keys.append(dense_key)
                        for dense_key in pair_dense_keys:
                            assert (
                                dialogs[dialog_key]["prompt_hash"]
                                == dialogs[dense_key]["prompt_hash"]
                            ), f"Prompt hash of {pair_dense_keys}, {dense_key} not equal!\n{dialogs[dialog_key]['messages']}\n{dialogs[dense_key]['messages']}"
                        fork_pairs.extend(
                            [(dialog_key, dense_key) for dense_key in pair_dense_keys]
                        )
            else:  # this tree has no dense, need to do outcome pairs
                for dialog_key in flattens:
                    dialog = dialogs[dialog_key]
                    if (dialog.get("operations") or [{}])[0].get("is_prompt_modified"):
                        # when prompt modified and no dense in this tree, don't need to as outcome negative supervision
                        break
                    for dense_key in dense_keys:
                        if dialog["prompt_hash"] == dialogs[dense_key]["prompt_hash"]:
                            outcome_pairs.append((dialog_key, dense_key))

        self.trees = trees
        self.to_parent = to_parent
        self.dense_keys = dense_keys
        self.outcome_pairs = outcome_pairs
        self.fork_pairs = fork_pairs
        self.valid_dialog_keys = dialog_valide_keys
        # g()

    def pre_process(self, data):
        assert "dialogs" in data, "invalid data format."
        data = deepcopy(data)
        data = recover_hash_map(data)
        assert self.SUPPORT_PANDA_TREE_VERSION >= data.get(
            "version", "0.0"
        ), f"Current parser support data version: {self.SUPPORT_PANDA_TREE_VERSION}, panda tree data version: {data['version']} Need to update onpanda package."
        assert (
            "update_time" in data
        ), "Never saved data. Which mean may never checked by Annotator."
        assert len(data["dialogs"]) >= 1, "Empty dialogs!"
        data["dialogs"] = {int(k): v for k, v in data["dialogs"].items()}
        # set default is_good
        max_key = max(data["dialogs"])
        for dialog_key in data["dialogs"]:
            dialog = data["dialogs"][dialog_key]
            if "annotate" not in dialog:
                dialog["annotate"] = {}
            if dialog["annotate"].get("is_good") is None:
                dialog["annotate"]["is_good"] = dialog_key == max_key

        dialog_valide_keys = [
            key
            for key in sorted(data["dialogs"].keys())
            if data["dialogs"][key]["messages"][-1]["role"] in RESPONSE_ROLES
        ]
        data["dialogs"] = {k: data["dialogs"][k] for k in dialog_valide_keys}
        self.prompt_hash_to_keys = {}
        for dialog_key in dialog_valide_keys:
            # set prompt_hash
            dialog = data["dialogs"][dialog_key]
            assert "annotate" in dialog, "No annotate in dialog!"
            prompt = mxlm.remove_last_assistant(dialog["messages"])
            dialog["prompt_hash"] = mxlm.hash_object_sha256_base64(prompt)
            self.prompt_hash_to_keys[dialog["prompt_hash"]] = (
                self.prompt_hash_to_keys.get(dialog["prompt_hash"], []) + [dialog_key]
            )
            dialog["sequence"] = self.messages_to_sequence(dialog["messages"])

            # set operations' parent to int
            for operation in dialog.get("operations", []):
                if "parent" in operation:
                    operation["parent"] = int(operation["parent"])
        return data

    def is_operation_tree_root(self, operations):
        if not operations:
            return True
        operation = operations[0]
        if operation.get("is_new_generated"):
            return True
        if operation.get("is_prompt_modified"):
            return True
        if not operation.get("parent"):
            return True

    def messages_to_sequence(self, messages):
        if not self.tokenizer:
            # default
            return mxlm.messages_to_sequence(messages)
        return self.tokenizer.apply_chat_template(messages, tokenize=False)

    def __str__(self):
        tree_str = str(self.trees).replace(": {}", ":''")
        s = f"""PandaTree({tree_str}):
    dense_keys: {self.dense_keys}
    fork_pairs: {self.fork_pairs}
    outcome_pairs: {self.outcome_pairs}"""
        return s

    __repr__ = __str__

    def build_legacy_data_v1(self, only_finish_reason_is_stop=False):
        data = self.data
        dialogs = data["dialogs"]
        sfts = []
        for dense_key in self.dense_keys:
            dialog = dialogs[dense_key]
            messages = deepcopy(dialog["messages"])
            # add onpanda
            onpanda_info = {"dialog_key": dense_key}
            if data.get("uuid"):
                onpanda_info["uuid"] = data["uuid"]
            messages[0]["onpanda"] = onpanda_info
            sfts.append(messages)
        preferences = []
        for rejected_key, chosen_key in self.outcome_pairs + self.fork_pairs:
            rejected = dialogs[rejected_key]
            chosen = dialogs[chosen_key]
            assert rejected["prompt_hash"] == chosen["prompt_hash"]
            preference = deepcopy(
                rejected["messages"][:-1]
                + [rejected["messages"][-1], chosen["messages"][-1]]
            )
            # key name 'chosen, rejected' from Anthropic/hh-rlhf
            preference[-1]["preference_tag"] = "chosen"
            preference[-2]["preference_tag"] = "rejected"
            onpanda_info = {"dialog_pair": (rejected_key, chosen_key)}
            onpanda_info["pair_type"] = (
                "fork" if (rejected_key, chosen_key) in self.fork_pairs else "outcome"
            )
            if data.get("uuid"):
                onpanda_info["uuid"] = data["uuid"]
            preference[0]["onpanda"] = onpanda_info
            if only_finish_reason_is_stop:
                if (
                    preference[-1].get("finish_reason") != "stop"
                    or preference[-2].get("finish_reason") != "stop"
                ):
                    continue
            preferences.append(preference)
        return dict(sfts=sfts, preferences=preferences)

    def build_token_level_supervision_data_v1(self, tokenizer=None):
        """
        token_level supervision data v1 structrue like this:
        tree-token_levels[0]
        └── /: list  2
            ├── 0: dict  2
            │   ├── role: user
            │   └── content: 1+1=? using English word
            └── 1: dict  4
                ├── role: assistant
                ├── content: list  3
                │   ├── 0: dict  3
                │   │   ├── type: text
                │   │   ├── text: 1+1=
                │   │   └── ignore_loss: True
                │   ├── 1: dict  4
                │   │   ├── type: text
                │   │   ├── text: two
                │   │   ├── ignore_loss: False
                │   │   └── tokens: list  1
                │   │       └── 0: 99473
                │   └── 2: dict  3
                │       ├── type: text
                │       ├── text: .
                │       └── ignore_loss: True
                ├── finish_reason: stop
                └── token_level: dict  4
                    ├── fork_token_idx: 15
                    ├── chosen_token_id: 99473
                    ├── rejected_token_id: 26288
                    ├── chosen_dialog_key: 3
                    ├── rejected_dialog_key: 2
                    ├── chosen_text: two
                    └── rejected_content: list  3
                        ├── 0: dict  3
                        │   ├── type: text
                        │   ├── text: 1+1=
                        │   └── ignore_loss: True
                        ├── 1: dict  5
                        │   ├── type: text
                        │   ├── text: three
                        │   ├── ignore_loss: False
                        │   ├── tokens: list  1
                        │   │   └── 0: 26288
                        │   └── rejected_loss: True
                        └── 2: dict  3
                            ├── type: text
                            ├── text: .
                            └── ignore_loss: True
        """
        tokenizer = tokenizer or self.tokenizer
        assert (
            tokenizer
        ), "token_level_supervision needs to set tokenizer, or you could using onpanda.unicode_tokenizer"
        token_levels = []
        for rejected_key, chosen_key in self.fork_pairs:
            chosen_msgs = self.data["dialogs"][chosen_key]["messages"]
            rejected_msgs = self.data["dialogs"][rejected_key]["messages"]
            chosen_content = chosen_msgs[-1]["content"]
            rejected_content = rejected_msgs[-1]["content"]
            token_level_info = compute_token_level_supervision(
                chosen_content=chosen_content,
                rejected_content=rejected_content,
                tokenizer=tokenizer,
            )
            token_level_info["version"] = "1.0"
            token_level_info["chosen_dialog_key"] = chosen_key
            token_level_info["rejected_dialog_key"] = rejected_key
            token_level_info["rejected_finish_reason"] = rejected_msgs[-1].get(
                "finish_reason", ""
            )
            token_level_msgs = chosen_msgs[:-1] + [
                {
                    **chosen_msgs[-1],
                    "content": token_level_info.pop("chosen_content"),
                    "token_level": token_level_info,
                }
            ]
            token_level_msgs = deepcopy(token_level_msgs)

            onpanda_info = {"dialog_pair": (rejected_key, chosen_key)}
            if self.data.get("uuid"):
                onpanda_info["uuid"] = self.data["uuid"]
            token_level_msgs[0]["onpanda"] = onpanda_info

            token_levels.append(token_level_msgs)
        # g()
        return token_levels

    def build_token_level_supervision_data_v2(self, tokenizer=None):
        """
        Merge multi token_level_supervision_data_v1 with same chosen_dialog_key into one messages with multi ignore_loss=False tokens (chosen tokens), thus to reduce SFT data number.
        Also gather token_level to token_levels

        Should save as 'xxx.token.json'
        """
        token_level_v1s = self.build_token_level_supervision_data_v1(
            tokenizer=tokenizer
        )
        chosen_dialog_key_to_token_level_v1s = {}
        for token_level_v1 in token_level_v1s:
            chosen_dialog_key = token_level_v1[-1]["token_level"]["chosen_dialog_key"]
            chosen_dialog_key_to_token_level_v1s.setdefault(
                chosen_dialog_key, []
            ).append(token_level_v1)
        token_level_v2s = []
        for chosen_dialog_key in sorted(chosen_dialog_key_to_token_level_v1s):
            msg = None
            ignore_loss_unicode_masks = []
            token_level_infos = []
            token_level_v1s = chosen_dialog_key_to_token_level_v1s[chosen_dialog_key]
            for token_level_v1 in sorted(
                token_level_v1s,
                key=lambda msgs: msgs[-1]["token_level"]["fork_token_idx"],
            ):
                if msg is None:
                    msg = deepcopy(token_level_v1[-1])
                    content_str0 = "".join([c["text"] for c in msg.pop("content")])
                    msg["token_levels"] = [msg.pop("token_level")]

                content = token_level_v1[-1]["content"]
                content_str = "".join([c["text"] for c in content])
                assert (
                    content_str == content_str0
                ), f"Chosen content string not equal: {content_str} != {content_str0}"
                ignore_loss_unicode_mask = sum(
                    [[c.get("ignore_loss", False)] * len(c["text"]) for c in content],
                    [],
                )
                ignore_loss_unicode_mask += [
                    content[-1].get("ignore_loss", False)
                ]  # concat stop token's mask
                ignore_loss_unicode_masks.append(ignore_loss_unicode_mask)
                token_level_infos.append(token_level_v1[-1]["token_level"])
            # ignore_loss_unicode_masks = np.array(ignore_loss_unicode_masks)
            # ignore_loss_unicode_mask_merged = ~(~ignore_loss_unicode_masks).any(0)
            ignore_loss_unicode_mask_merged = [
                all(row[i] for row in ignore_loss_unicode_masks)
                for i in range(len(ignore_loss_unicode_masks[0]))
            ]
            chosen_content_merged = apply_ignore_unicode_loss_mask_to_content(
                ignore_loss_unicode_mask_merged, content_str0
            )
            token_level_v2 = deepcopy(token_level_v1)
            token_level_v2[-1]["content"] = chosen_content_merged
            token_level_v2[-1]["token_levels"] = token_level_infos
            token_level_v2[-1].pop("token_level")
            token_level_v2s.append(token_level_v2)
        # g() / 0
        return token_level_v2s

    def build_correcting_sft_data_v1(self, ntp_as_correcting_builder):
        """
        tree(correcting_sfts[-1])
        ├── 1: dict  3
        │   ├── role: user
        │   ├── content: How many 1 in 01011010101111011011?
        │   └── description: Answer is 13
        ├── 2: dict  5
        │   ├── role: assistant
        │   ├── ignore_loss: True
        │   ├── content: To determine how many times the digit 1 ap...
        │   ├── finish_reason: stop
        │   └── token_level: dict  11
        ├── 3: dict  2
        │   ├── role: system
        │   └── content: {correcting SFT system prompt}...
        └── 4: dict  3
            ├── role: assistant
            ├── content: <|fim_pad|> **0<|fim_pad|>0<|fim_pad|> <|f...
            └── correcting: dict  7
                ├── location_text:  **0
                ├── match_num: 1
                ├── location_index: 0
                ├── location_tokens: list  2
                │   ├── 0: 3070
                │   └── 1: 15
                ├── replacement_text:
                ├── is_good: False
                └── scope_slice: tuple 2
                    ├── 0: -1
                    └── 1: None
        """
        sfts = self.build_legacy_data_v1()["sfts"]
        correcting_sfts = [
            ntp_as_correcting_builder.build_correcting_sft_by_token_level_SFT(
                sft, is_good=True
            )
            for sft in sfts
        ]
        token_level_v1s = self.build_token_level_supervision_data_v1(
            tokenizer=ntp_as_correcting_builder.tokenizer
        )
        correcting_sfts += [
            ntp_as_correcting_builder.build_correcting_sft_by_token_level_SFT(
                sft, is_good=False
            )
            for sft in token_level_v1s
        ]
        return correcting_sfts


def build_test_panda_tree(panda_json=None, tokenizer=None):
    if panda_json is None:
        panda_json = "../../on-panda-example-data/panda_json/2025-08-19_how-many-1s_tokenizer-Qwen2.5.panda.json"
    if isinstance(panda_json, str):
        panda_json = json.load(open(panda_json))
    panda_tree = PandaTree(panda_json, tokenizer=tokenizer)
    return panda_tree


if __name__ == "__main__":
    from boxx import *  # pip install boxx

    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    # tokenizer = None
    tokenizer = __import__("transformers").AutoTokenizer.from_pretrained(
        "Qwen/Qwen2.5-7B-Instruct-GPTQ-Int4"
    )
    # %%

    # tokenizer = unicode_tokenizer
    # test_json = "../../asset/on-panda-example/shape-of-V-test-hash.panda.json"
    # test_json = "../../asset/on-panda-example/how-many-1s.panda.json"
    # test_json = "../../on-panda-example-data/panda_json/2025-04-12_Chinese_acrostic_poem_藏头诗_tokenizer-step2.panda.json"
    # test_json = "../../on-panda-example-data/panda_json/2025-09-10_correcting_sft_tokenizer-Qwen2.5.panda.json"
    test_json = "../../on-panda-example-data/panda_json/2025-08-19_how-many-1s_tokenizer-Qwen2.5.panda.json"

    panda_tree = build_test_panda_tree(test_json, tokenizer)
    print(panda_tree)
    tree(panda_tree.trees)

    legacy_data = panda_tree.build_legacy_data_v1()
    sfts = legacy_data["sfts"]
    # tree(legacy_data)

    token_level_v1s = panda_tree.build_token_level_supervision_data_v1(
        tokenizer=tokenizer
    )

    token_level_v2s = panda_tree.build_token_level_supervision_data_v2(
        tokenizer=tokenizer
    )
    sft_correcting_builder = NextTokenPredictionAsCorrectingBuilder(
        tokenizer=tokenizer,
        SPLIT_TOKEN="<|fim_pad|>",  # for qwen 2.5
        STOP_TOKEN="<|fim_suffix|>",
        max_location_tokens=20,
    )

    correcting_sfts = panda_tree.build_correcting_sft_data_v1(sft_correcting_builder)

    tree(correcting_sfts[-1])
    # savejson(correcting_sfts[-1], "/home/yl/onPanda/asset/correcting_sft/correcting_sft_example1.sft.json")
    # tree(correcting_sfts[3])



def build_test_tokenizer(name_or_path="Qwen/Qwen2.5-7B-Instruct-GPTQ-Int4"):
    from transformers import AutoTokenizer

    try:
        tokenizer = AutoTokenizer.from_pretrained(
            name_or_path,
            use_fast=True,
            local_files_only=True,
        )
    except (OSError, ValueError):
        # Smaller model that ships with Transformers – guarantees the
        # example runs even without the Llama‑3 files.
        tokenizer = AutoTokenizer.from_pretrained("gpt2")
    return tokenizer


def get_test_rejected_msgs1():
    rejected_msgs1 = [
        {"role": "user", "content": "Name three kinds of fruit:"},
        {
            "role": "assistant",
            "content": "Apple, potato, banana.",
            "finish_reason": "stop",
            "token_level": {
                "chosen_text": " orange",
                "rejected_text": " potato",
                "chosen_text_unicode_range": [6, 13],  # " potato" start at index 6
                "rejected_text_unicode_range": [6, 13],
                "version": "1.0",
                "chosen_dialog_key": 2,
                "rejected_dialog_key": 1,
                "rejected_finish_reason": "stop",
            },
        },
    ]
    ntp_as_correcting_text_gt1 = (
        "<|fim_pad|> potato<|fim_pad|>0<|fim_pad|> orange<|fim_pad|>"
    )
    return rejected_msgs1, ntp_as_correcting_text_gt1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 23 16:56:53 2025

@author: DIYer22
"""


class IsGoodScoreMixin:
    def compute_is_good_score(
        self,
        messages,
    ):
        # add correcting SFT system prompt and is_good answer (`<|split|><|split|>`)
        is_good_msgs = self.builder.build_correcting_sft_by_token_level_SFT(
            messages, is_good=True
        )
        # prefill prmpt_logprobs to get is_good probability
        dic = self.chat_correcting(
            is_good_msgs,
            return_dict=True,
            max_tokens=1,
            temperature=1.0,
            top_p=1.0,
            logprobs=True,
            # top_logprobs=1,
            extra_body=dict(
                prompt_logprobs=True,
                add_generation_prompt=False,
                continue_final_message=True,
                skip_special_tokens=False,
            ),
        )
        first_split_token = list(dic["prompt_logprobs"][-2].values())[0]
        assert (
            first_split_token["decoded_token"] == self.builder.SPLIT_TOKEN
        ), f"first_split_tokens: {dic['prompt_logprobs'][-2]}, self.builder.SPLIT_TOKEN: {self.builder.SPLIT_TOKEN}"
        e = 2.718281828459045  # base of the natural logarithm
        prob_first_split = e ** first_split_token["logprob"]
        assert (
            prob_first_split > 0.99
        ), f"CorrectingSftModel should learn to output `{self.builder.SPLIT_TOKEN}` first. first_split_tokens: {dic['prompt_logprobs'][-2]}, prob_first_split: {prob_first_split}"

        second_split_token = list(dic["prompt_logprobs"][-1].values())[0]
        assert (
            second_split_token["decoded_token"] == self.builder.SPLIT_TOKEN
        ), f"second_split_tokens: {dic['prompt_logprobs'][-1]}, self.builder.SPLIT_TOKEN: {self.builder.SPLIT_TOKEN}"
        is_good_logprob = second_split_token["logprob"]
        is_good_prob = e**is_good_logprob
        is_good_score = dict(is_good_prob=is_good_prob, is_good_logprob=is_good_logprob)

        if "using chat_correcting.prefill_logprobs for double check" and 0:
            prefill_logprobs = self.chat_correcting.prefill_logprobs(is_good_msgs)[-1][
                "prefill_logprobs"
            ]
            tree - prefill_logprobs
            print(is_good_prob)
            is_good_prob = e ** sum([d["logprob"] for d in prefill_logprobs])
            print(is_good_prob)
            g()
        return is_good_score


if __name__ == "__main__":
    from boxx import *
    import transformers
    import mximport

    with mximport.inpkg():
        from .correcting_sft_model import build_test_correcting_sft_model

    correct_model = build_test_correcting_sft_model()

    msgs = [
        {"role": "user", "content": "5+7="},
        {"role": "assistant", "content": "32"},
        # {"role": "assistant", "content": "12"},
    ]
    msgs = [
        {"role": "user", "content": "Name three kinds of fruit:"},
        {
            "role": "assistant",
            "content": "Apple, potato, banana.",
            # "content": "Apple, orange, banana.",
        },
    ]

    is_good_score = correct_model.compute_is_good_score(msgs)
    print(f'{is_good_score["is_good_prob"]*100:04.1f}%')



from typing import Union

from fastapi import FastAPI, APIRouter

FILE_SUFFIX = ".panda.json"

app = FastAPI()


@app.get("/")
def read_root():
    return {"Hello": "onPanda"}


@app.get("/items/{item_id}")
def read_item(item_id: int, q: Union[str, None] = None):
    return {"item_id": item_id, "q": q}


router = APIRouter(prefix="/panda-api/v1")


@router.get("/")
def read_root():
    FILE_SUFFIX
    return {"Hello": "onPanda"}


app.include_router(router)
